import copy

from .game_settings import P2_MARKER, P2_WON, P1_WON, DRAW_GAME, OCCUPIED_POSITION, MARK_WAS_PLACED, P1_MARKER


class Game:
    """
       Game logic for a 2 player tic-tac-toe game.

       Attributes:
           game_is_on (bool): Flag that indicates whether a game is been played. Turns off after a victory or a draw.
           is_player1_turn (bool): Flag that controls players turn. First to play is randomly generated.
           total_moves (int): Keeps the number of moves played. Used to check for game ties.
           board_state (list): 3 * 3 array that keeps the game board current layout.
           position_x (int): Keeps the cursor x position on the game board.
           position_y (int): Keeps the cursor y position on the game board.
           plays (array): Keeps all the board_states in a game.
       """

    def __init__(self):
        self.game_is_on = True
        self.is_player1_turn = True
        self.total_moves = 0
        self.board_state = [list('   ') for _ in range(3)]
        self.position_x = 1
        self.position_y = 1
        self.plays = []

    def end_game(self):

        """
        Ends current game.

        Called when a player wins or when a tie is reached.
        :return: Nothing
        """

        self.game_is_on = False

    def player_input(self, pressed_key):
        """

        :param pressed_key: string representation of pressed_key.
        :return: string outcome of pressed key (check game_settings.py)
        """

        if pressed_key != ' ':
            self.move_cursor_to_position(pressed_key)
            return None
        else:
            if self.is_position_empty():

                self.place_marker()
                self.increase_total_moves()
                self.save_board_state()

                if self.total_moves > 4:  # waits to the fifth move to start checking for victories
                    if self.has_player_won():
                        return P1_WON if self.is_player1_turn else P2_WON
                    elif self.total_moves > 8:
                        return DRAW_GAME

                self.change_player_turn()

                return MARK_WAS_PLACED
            else:
                return OCCUPIED_POSITION

    def save_board_state(self):
        self.plays.append(copy.deepcopy(self.board_state))

    def change_player_turn(self):
        self.is_player1_turn = not self.is_player1_turn

    def increase_total_moves(self):
        self.total_moves += 1

    def is_position_empty(self):
        """

        :return: bool indicating whether the slot is empty
        """
        y = self.position_y
        x = self.position_x
        board = self.board_state
        return board[y][x] == ' '

    def move_cursor_to_position(self, move_order):
        """
        keeps track of the cursor position on the board
        :param move_order: string representation of pressed key
        :return:
        """
        if move_order == "KEY_UP":
            self.position_y = max(0, self.position_y - 1)
        elif move_order == "KEY_DOWN":
            self.position_y = min(2, self.position_y + 1)
        elif move_order == "KEY_LEFT":
            self.position_x = max(0, self.position_x - 1)
        elif move_order == "KEY_RIGHT":
            self.position_x = min(2, self.position_x + 1)

    def place_marker(self):

        """
        Marks the board with current player mark.

        :return: Nothing
        """
        y = self.position_y
        x = self.position_x
        board = self.board_state

        board[y][x] = P1_MARKER if self.is_player1_turn else P2_MARKER

    def has_player_won(self):
        """
        Checks, after each play, if someone has won.

        :return: True if a winning pattern is present on the board. False otherwise.
        """

        y = self.position_y
        x = self.position_x
        board = self.board_state

        return (
                (board[0][x] == board[1][x] == board[2][x])  # horizontal victory
                or (board[y][0] == board[y][1] == board[y][2])  # vertical victory
                or (y == x and board[0][0] == board[1][1] == board[2][2])  # main diagonal
                or (y + x == 2 and board[0][2] == board[1][1] == board[2][0])  # secondary diagonal
        )
